@@ -1,10 +1,11 @@
 /* xxkb  - XKB keyboard indicator/switcher */
-/* (c)  1999 2000 Ivan Pascal <pascal@tsu.ru>   */
+/* (c)  1999 - 2003 Ivan Pascal <pascal@tsu.ru>   */
 
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/XKBlib.h>
 #include <X11/Xlibint.h>
+#include <X11/Xatom.h>
 
 #include "wlist.h"
 #include "xxkb.h"
@@ -29,7 +30,7 @@ XWMHints *wm_hints;
 XSizeHints *size_hints;
 XClassHint class_hints;
 char *name=NAME;
-Atom take_focus_atom;
+Atom take_focus_atom, wm_del_win;
 
 WInfo def_info, *info, *tmp_info;
 kbdState def_state;
@@ -40,19 +41,42 @@ static ListAction  FindAppList(Window w);
 
 int main (int argc, char ** argv)
 {
-  int  xkbEventType, xkbError, reason_rtrn;
+  int  xkbEventType, xkbError, reason_rtrn, mjr, mnr;
   Bool fout_flag = False;
   Geometry geom;
-
+  Atom protocols[1];
+  
 /* Lets begin */
   dpy = XkbOpenDisplay( "", &xkbEventType, &xkbError,
 			  NULL, NULL, &reason_rtrn); 
-  if (!dpy) {puts("Can' connect to X-server"); exit(-1);}
-    
+  if (!dpy) {
+    printf("Can't connect to X-server: %s\n", getenv("DISPLAY"));
+    switch (reason_rtrn) {
+       case XkbOD_BadLibraryVersion :
+       case XkbOD_BadServerVersion :
+            printf("xxkb was compiled with XKB version %d.%02d\n",
+                    XkbMajorVersion,XkbMinorVersion);
+            printf("But %s uses incompatibl eversion %d.%02d\n",
+                   reason_rtrn == XkbOD_BadLibraryVersion ? "Xlib" : "Xserver",
+                   mjr,mnr);
+            break;
+      case XkbOD_ConnectionRefused :
+           printf("Connection refused\n");
+           break;
+      case XkbOD_NonXkbServer:
+           printf("XKB extension not present\n");
+           break;
+      default:
+           printf("Unknown error %d from XkbOpenDisplay\n", reason_rtrn);
+      }
+    exit(1);
+  }    
+
   scr = DefaultScreen(dpy);
   root = RootWindow(dpy, scr);
   base_mask = ~(dpy->resource_mask);
   take_focus_atom = XInternAtom(dpy, "WM_TAKE_FOCUS", True);
+  wm_del_win = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
   DefErrHandler = XSetErrorHandler((XErrorHandler) ErrHandler);
  
   focused_event.type = FocusIn;
@@ -105,6 +129,9 @@ int main (int argc, char ** argv)
     XSetNormalHints(dpy, MainWin, size_hints);
   }
 
+  *protocols = wm_del_win;
+  XSetWMProtocols(dpy, MainWin, protocols, 1);
+
 /* Show window ? */
   if (conf.controls&WMaker) {
     icon = XCreateSimpleWindow( dpy, MainWin,
@@ -117,8 +144,27 @@ int main (int argc, char ** argv)
     wm_hints->initial_state = WithdrawnState;
     wm_hints->flags = wm_hints->flags | StateHint | IconWindowHint;
     XSetWMHints(dpy, MainWin, wm_hints);
+
   }
-  else icon = NULL;
+  else icon = (Window) 0;
+
+    if (conf.tray_type) {
+       Atom r;
+       int data = 1;
+       if (! strcmp(conf.tray_type, "KDE") ||
+           ! strcmp(conf.tray_type, "GNOME") ) {
+
+           r = XInternAtom(dpy, "KWM_DOCKWINDOW", False);
+           XChangeProperty(dpy, MainWin, r, r, 32, 0,
+                           (unsigned char *)&data, 1);
+       } else if (! strcmp(conf.tray_type, "KDE2")) {
+           r = XInternAtom(dpy, "_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR", False);
+           XChangeProperty(dpy, MainWin, r, XA_WINDOW, 32, 0,
+                           (unsigned char *)&data, 1);
+       }
+    }
+
+
   
   if (conf.controls&Main_enable) XMapWindow(dpy, MainWin);
 
@@ -144,16 +190,16 @@ int main (int argc, char ** argv)
   Reset();
 
   if (conf.controls&When_start) {
-    Window rwin, parent, *childrens, *child, app; int num; 
-    XQueryTree(dpy, root, &rwin, &parent, &childrens, &num);
-    child = childrens;
+    Window rwin, parent, *children, *child, app; int num; 
+    XQueryTree(dpy, root, &rwin, &parent, &children, &num);
+    child = children;
     while (num) {
-      app = NULL;
+      app = (Window) 0;
       GetAppWindow(*child, &app);
       if (app) AddWindow(app, *child);
       child++; num--;
     }
-    if (childrens) XFree(childrens);
+    if (children) XFree(children);
     XGetInputFocus(dpy, &focused, &revert);
     info = win_find(focused);
     if (!info) info = &def_info;
@@ -236,32 +282,45 @@ int main (int argc, char ** argv)
 	            XkbLockGroup(dpy, XkbUseCoreKbd, conf.Base_group);
 	        }
 	        else
-	          XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group + 1);
+                  if (conf.controls&But1_reverse) {
+	            XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group - 1);
+                  } else {
+	            XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group + 1);
+                  }
 	      }
 	      break;
 	    case Button3:
 	      if ((win == info->button) || (win == MainWin) ||
 		  (icon && (win == icon))) {
-	        XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group + 1);
+                if (conf.controls&But3_reverse) {
+	           XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group - 1);
+	        } else {
+	           XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group + 1);
+	        }
 	      }
 	      break;
 	    case Button2:
 	      if ((win != MainWin) && (win != icon)){
 	      	if (conf.controls&Button_delete) {
-	      	  XDestroyWindow(dpy,win);
+	      	  XDestroyWindow(dpy, win);
+                  if (conf.controls&Forget_window) {
+                     if (win == info->button) {
+                        AddAppToConfig(info->win, ev.core.xbutton.state);
+	                info->ignore = 1;
+	                Reset();
+                     } else {
+                        tmp_info = button_find(win);
+	                if (tmp_info) {
+                           AddAppToConfig(tmp_info->win, ev.core.xbutton.state);
+	                   tmp_info->ignore = 1;
+	                }
+                     }
+                  }
 	      	}
 	        break;
 	      }
-	      win_free_list();
-	      XFreeGC(dpy,gc);
-	      if (icon) XDestroyWindow(dpy, icon);
-	      XDestroyWindow(dpy,MainWin);
-#ifdef XT_RESOURCE_SEARCH
-	      XtCloseDisplay(dpy);
-#else
-	      XCloseDisplay(dpy);
-#endif
-	      exit(0);
+              if (conf.controls&Main_delete)
+                Terminate();
 	  }
 	  break;
 	case FocusIn:
@@ -270,7 +329,12 @@ int main (int argc, char ** argv)
 	    printf ("Oops. FocusEvent from unknown window\n");
 	    info = &def_info;
 	  }
-	  XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group);
+	  if (!info->ignore) {
+	    XkbLockGroup(dpy, XkbUseCoreKbd, info->state.group);
+	  } else {
+            def_info.state.group = info->state.group;
+            info = &def_info;
+	  }
 	  break;
 	case FocusOut:
 	  if (conf.controls&Focus_out) {
@@ -313,6 +377,21 @@ int main (int argc, char ** argv)
           if (tmp_info)
              XRaiseWindow(dpy, ev.core.xconfigure.above);
           break;
+        case PropertyNotify:
+          win = ev.core.xproperty.window;
+	  tmp_info = win_find(win);
+          if (!tmp_info) {
+             Window rwin, parent, *childrens, *child, app; int num;
+             XQueryTree(dpy, win, &rwin, &parent, &childrens, &num);
+             AddWindow(win, parent);
+          }
+          break;
+        case ClientMessage:
+          win = ev.core.xclient.window;
+          if(((win == MainWin) || (win == icon))
+             && ev.core.xclient.data.l[0] == wm_del_win)
+             Terminate();
+          break;
         case CreateNotify:
         case UnmapNotify:
         case MapNotify:
@@ -330,9 +409,24 @@ int main (int argc, char ** argv)
 void Reset()
 {
   info = &def_info;
+  info->state = def_state;
   XkbLockGroup(dpy, XkbUseCoreKbd, conf.Base_group);
 }
 
+void Terminate()
+{
+  win_free_list();
+  XFreeGC(dpy,gc);
+  if (icon) XDestroyWindow(dpy, icon);
+  XDestroyWindow(dpy,MainWin);
+#ifdef XT_RESOURCE_SEARCH
+  XtCloseDisplay(dpy);
+#else
+  XCloseDisplay(dpy);
+#endif
+  exit(0);
+}
+
 void getGC(win, gc)
     Window win;
     GC * gc;
@@ -362,42 +456,50 @@ Window win; GC gc; int group;
 WInfo* AddWindow(win, parent)
 Window win, parent;
 {
+  int ignore = 0;
   WInfo *info;
   ListAction action;
   XWindowAttributes attr;
 
+  /* properties can be unappropriate at this moment */
+  /* so we need posibility to redecide when them will be changed*/
+  XSelectInput(dpy, win, PropertyChangeMask);
+
+  /* don't deal with windows that never get a focus */
   if (!ExpectInput(win))
-     return NULL;
+     return (WInfo*) 0;
 
   action = FindAppList(win);
   if ( ((action&Ignore) && !(conf.controls&Ignore_reverse)) ||
        (!(action&Ignore) && (conf.controls&Ignore_reverse)))
-     return NULL;
+     ignore = 1;
 
   info = win_find(win);
   if (!info) {
     info = win_add(win, &def_state);
-    if (!info) return NULL;
+    if (!info) return (WInfo*) 0;
     if (action&AltGrp) {
        info->state.alt = action & GrpMask;
     }
     if (action&InitAltGrp) {
        info->state.group = info->state.alt;
     }
-    XGetInputFocus(dpy, &focused, &revert);
-    XSelectInput(dpy, win, FocusChangeMask | StructureNotifyMask);
-    if (focused == win) {
-       focused_event.window = win;
-       XPutBackEvent(dpy, (XEvent *) &focused_event);
-    }
   }
-  if ((conf.controls&Button_enable) && (!info->button))
+  XGetInputFocus(dpy, &focused, &revert);
+  XSelectInput(dpy, win, FocusChangeMask | StructureNotifyMask |  PropertyChangeMask);
+  if (focused == win) {
+      focused_event.window = win;
+      XSendEvent(dpy, MainWin, 0, 0, (XEvent *) &focused_event);
+  }
+
+  info->ignore = ignore;
+  if ((conf.controls&Button_enable) && (!info->button) && !ignore )
     info->button = MakeButton(parent);
 
 /* to be sure that window still exists */
   if(!XGetWindowAttributes(dpy, win, &attr)) {
      win_free(win);
-     return NULL;
+     return (WInfo*) 0;
   }
   return info;
 }
@@ -409,11 +511,10 @@ Window parent;
   int x, y; unsigned int width, height, bord, dep;
   XSetWindowAttributes attr;
   Geometry geom = conf.but_geom;
-  if((parent = GetGrandParent(parent)) == NULL)
-     return NULL;
+  if((parent = GetGrandParent(parent)) == (Window) 0)
+     return (Window) 0;
 
-  if (geom.mask & (XNegative|YNegative))
-    XGetGeometry(dpy, parent, &rwin,
+  XGetGeometry(dpy, parent, &rwin,
 		 &x, &y, &width, &height, &bord, &dep);
   x = (geom.mask & XNegative)? width  + geom.x - geom.width  : geom.x;
   y = (geom.mask & YNegative)? height + geom.y - geom.height : geom.y;
@@ -440,7 +541,7 @@ Window w;
 
   while (1) {
      if(! XQueryTree(dpy, w, &rwin, &parent, &child, &num))
-        return NULL;
+        return (Window) 0;
      if (child) XFree(child);
      if (parent == rwin) return w;
      w = parent;
@@ -512,6 +613,47 @@ ListAction searchInList(SearchList *list, char *str)
     return ret;
 }
 
+static
+ListAction searchInPropList(SearchList *list, Window win)
+{
+    int i, j;
+    int prop_num;
+    Atom *props, *atoms;
+    ListAction ret = 0;
+
+    while (list) {
+       atoms = (Atom *) calloc(list->num, sizeof(Atom));
+       if (!atoms)
+           break;
+       XInternAtoms(dpy, list->idx, list->num, False, atoms);
+       j = 0;
+       for (i = 0; i < list->num; i++) {
+           if (atoms[i] != None) {
+              j = 1;
+              break;
+           }
+       }
+       if (j != 0) {
+          props = XListProperties(dpy, win, &prop_num);
+          if (props) {
+              for (i = 0; i < list->num; i++) {
+                  if (atoms[i] != None) {
+                      for (j = 0; j < prop_num; j++) {
+                         if (atoms[i] == props[j])
+                             ret |= list->action;
+                      }
+	          }
+              }
+              XFree(props);
+          }
+       }
+       XFree(atoms);
+       list = list->next;
+    }
+    return ret;
+}
+
+
 static
 ListAction FindAppList(Window w)
 {
@@ -532,6 +674,8 @@ ListAction FindAppList(Window w)
         Xfree(name);
      }
   }
+  ret |= searchInPropList(conf.lists[3], w);
+
   return ret;
 }
 
@@ -563,6 +707,57 @@ Bool ExpectInput(Window w)
   return ok;
 }
 
+char* GetAppIdent(Window appwin, MatchType type)
+{
+  XClassHint wm_class;
+  char *name;
+
+  if (type == WMName) {
+     if(XFetchName(dpy, appwin, &name)) {
+     	return name;
+     } else {
+     	return NULL;
+     }
+  }
+  if( XGetClassHint(dpy, appwin, &wm_class)){
+     if (type == WMClassClass) {
+        Xfree(wm_class.res_name);
+     	return wm_class.res_class;
+     } else {
+        Xfree(wm_class.res_class);
+     	return wm_class.res_name;
+     }
+  } else {
+     return NULL;
+  }
+}
+
+void AddAppToConfig(Window appwin, unsigned int state)
+{
+   char *name = NULL;
+   MatchType type;
+
+   switch (state & (ControlMask | ShiftMask)) {
+      case 0:
+         return;
+      case ControlMask :
+         type = WMClassClass;
+         break;
+      case ShiftMask   :
+         type = WMName;
+         break;
+      case ControlMask | ShiftMask:
+         type = WMClassName;
+   } 
+
+   if ((name = GetAppIdent(appwin, type)) != NULL) {
+       SaveAppInConfig(&conf, name, type);
+       Xfree(name);
+   } else {
+       XBell(dpy, conf.Bell_percent);
+   }
+}
+
 void ErrHandler(Display *dpy, XErrorEvent* err)
 {
   if((err->error_code == BadWindow) || (err->error_code == BadDrawable))
